/*
 * Copyright (C) 2004-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file ogg_afh.c Audio format handler for ogg vorbis files. */

#include <inttypes.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>

#include "para.h"
#include "afh.h"
#include "error.h"
#include "string.h"

/* Taken from decoder_example.c of libvorbis-1.2.3. */
static int read_vorbis_comment(ogg_sync_state *oss, ogg_stream_state *stream,
		vorbis_info *vi, vorbis_comment *vc)
{
	ogg_page page;
	ogg_packet packet;
	int i = 0;

	while (i < 2) {
		while (i < 2) {
			int ret = ogg_sync_pageout(oss, &page);
			if (ret == 0)
				break; /* Need more data */
			if (ret != 1)
				continue;
			/*
			 * We can ignore any errors here as they'll also become
			 * apparent at packetout.
			 */
			ogg_stream_pagein(stream, &page);
			while (i < 2) {
				ret = ogg_stream_packetout(stream, &packet);
				if (ret == 0)
					break;
				if (ret < 0)
					return -E_STREAM_PACKETOUT;
				ret = vorbis_synthesis_headerin(vi, vc,
					&packet);
				if (ret < 0)
					return -E_VORBIS;
				i++;
			}
		}
	}
	return 1;
}

static int read_vorbis_info(ogg_sync_state *oss, struct afh_info *afhi)
{
	vorbis_comment vc;
	vorbis_info vi;
	ogg_packet packet;
	ogg_stream_state stream;
	ogg_page page;
	int ret;
	char *taginfo;

	vorbis_info_init(&vi);
	vorbis_comment_init(&vc);

	ret = -E_SYNC_PAGEOUT;
	if (ogg_sync_pageout(oss, &page) != 1)
		goto out;

	ret = ogg_page_serialno(&page);
	ogg_stream_init(&stream, ret);

	ret = -E_STREAM_PAGEIN;
	if (ogg_stream_pagein(&stream, &page) < 0)
		goto out;

	ret = -E_STREAM_PACKETOUT;
	if (ogg_stream_packetout(&stream, &packet) != 1)
		goto out;

	ret = -E_VORBIS;
	if (vorbis_synthesis_headerin(&vi, &vc, &packet) < 0)
		goto out;
	if (vi.rate == 0)
		goto out;
	afhi->channels = vi.channels;
	afhi->frequency = vi.rate;
	afhi->bitrate = vi.bitrate_nominal / 1000;
	PARA_DEBUG_LOG("channels: %i, sampling rate: %i, bitrate: %i\n",
		afhi->channels, afhi->frequency, afhi->bitrate);
	ret = read_vorbis_comment(oss, &stream, &vi, &vc);
	if (ret < 0)
		goto out;
	taginfo = make_taginfo(
		vorbis_comment_query(&vc, "title", 0),
		vorbis_comment_query(&vc, "artist", 0),
		vorbis_comment_query(&vc, "album", 0),
		vorbis_comment_query(&vc, "year", 0),
		vorbis_comment_query(&vc, "comment", 0)
	);
	PARA_DEBUG_LOG("tag info: %s\n", taginfo);
	afhi->info_string = make_message("%s:\n%s",
		status_item_list[SI_AUDIO_FILE_INFO], taginfo);
	free(taginfo);

	afhi->header_offset = 0;
	afhi->header_len = oss->returned;
	ret = 1;
out:
	vorbis_info_clear(&vi);
	vorbis_comment_clear(&vc);
	ogg_stream_clear(&stream);
	return ret;
}

static void set_chunk_tv(int num_frames, int num_chunks, int frequency,
		struct timeval *result)
{
	uint64_t x = (uint64_t)num_frames * 1000 * 1000
		/ frequency / num_chunks;

	result->tv_sec = x / 1000 / 1000;
	result->tv_usec = x % (1000 * 1000);
	PARA_INFO_LOG("%d chunks, chunk time: %lums\n", num_chunks,
		tv2ms(result));
}

/* Write tech data to given audio format handler struct. */
static int ogg_get_file_info(char *map, size_t numbytes, __a_unused int fd,
		struct afh_info *afhi)
{
	ogg_sync_state oss;
	ogg_page op;
	long len = numbytes;
	char *buf;
	int ret, i, j, frames_per_chunk, ct_size;
	long long unsigned num_frames = 0;

	afhi->info_string = NULL;
	ogg_sync_init(&oss);
	ret = -E_OGG_SYNC;
	buf = ogg_sync_buffer(&oss, len);
	if (!buf)
		goto out;
	memcpy(buf, map, len);
	ret = -E_OGG_SYNC;
	if (ogg_sync_wrote(&oss, len) < 0)
		goto out;
	ret = read_vorbis_info(&oss, afhi);
	if (ret < 0)
		goto out;
	oss.returned = 0;
	oss.fill = numbytes;
	/* count ogg packages and get duration of the file */
	for (i = 0; ogg_sync_pageseek(&oss, &op) > 0; i++)
		num_frames = ogg_page_granulepos(&op);
	PARA_INFO_LOG("%d pages, %llu frames\n", i, num_frames);
	ret = -E_OGG_EMPTY;
	if (i == 0)
		goto out;
	afhi->seconds_total = num_frames / afhi->frequency;
	/* use roughly one page per chunk */
	frames_per_chunk = num_frames / i;
	PARA_INFO_LOG("%lu seconds, %d frames/chunk\n",
		afhi->seconds_total, frames_per_chunk);
	ct_size = 250;
	afhi->chunk_table = para_malloc(ct_size * sizeof(uint32_t));
	afhi->chunk_table[0] = 0;
	afhi->chunk_table[1] = afhi->header_len;
	oss.returned = afhi->header_len;
	oss.fill = numbytes;
	for (i = 0, j = 1; ogg_sync_pageseek(&oss, &op) > 0; i++) {
		int granule = ogg_page_granulepos(&op);

		while (granule > j * frames_per_chunk) {
			j++;
			if (j >= ct_size) {
				ct_size *= 2;
				afhi->chunk_table = para_realloc(
					afhi->chunk_table,
					ct_size * sizeof(uint32_t));
			}
			afhi->chunk_table[j] = oss.returned;
		}
	}
	afhi->chunks_total = j;
	set_chunk_tv(num_frames, j, afhi->frequency, &afhi->chunk_tv);
	tv_scale(3, &afhi->chunk_tv, &afhi->eof_tv);
	ret = 0;
out:
	ogg_sync_clear(&oss);
	return ret;
}

static const char* ogg_suffixes[] = {"ogg", NULL};

/**
 * The init function of the ogg vorbis audio format handler.
 *
 * \param afh Pointer to the struct to initialize.
 */
void ogg_init(struct audio_format_handler *afh)
{
	afh->get_file_info = ogg_get_file_info,
	afh->suffixes = ogg_suffixes;
}
